from typing import List

import z3
import AF
import z3_instance as zi


class Semantics:
    def __init__(self, z3i: zi.Z3Instance):
        self.context = z3i.context
        self.af = z3i.af
        self.z3_instance = z3i
        self.constraints: List[z3.BoolRef] = []

        self.generate_constraints()

    def generate_constraints(self):
        # implemented by subclasses
        pass

    def get_constraints(self):
        return self.constraints


####################
# trivial semantics
####################

class SemanticsMin(Semantics):
    """Minimality semantics: All nodes must hold with probability 0."""
    def generate_constraints(self):
        for node in self.af.get_nodes():
            self.constraints.append(node.prob_var == 0)


class SemanticsNeu(Semantics):
    """Neutrality semantics: All nodes must hold with probability 0.5."""
    def generate_constraints(self):
        for node in self.af.get_nodes():
            self.constraints.append(node.prob_var == 0.5)


class SemanticsMax(Semantics):
    """Maximality semantics: All nodes must hold with probability 1."""
    def generate_constraints(self):
        for node in self.af.get_nodes():
            self.constraints.append(node.prob_var == 1)


class SemanticsDirac(Semantics):
    """Dirac semantics: All nodes must hold with either probability 0 or 1."""
    def generate_constraints(self):
        for node in self.af.get_nodes():
            self.constraints.append(z3.Or(node.prob_var == 0, node.prob_var == 1))


class SemanticsTer(Semantics):
    """Ternary semantics: All nodes must hold with either probability 0, 0.5 or 1."""
    def generate_constraints(self):
        for node in self.af.get_nodes():
            self.constraints.append(z3.Or(node.prob_var == 0, node.prob_var == 0.5, node.prob_var == 1))


################################
# semantics by Hunter and Thimm
################################

class SemanticsFou(Semantics):
    """Foundedness semantics: Initial nodes must hold with probability 1."""
    def generate_constraints(self):
        for node in self.af.get_initial_nodes():
            self.constraints.append(node.prob_var == 1)


class SemanticsSFou(Semantics):
    """Semi-Foundedness semantics: Initial nodes must hold with probability >= 0.5."""
    def generate_constraints(self):
        for node in self.af.get_initial_nodes():
            self.constraints.append(node.prob_var >= 0.5)


class SemanticsInv(Semantics):
    """Involution semantics: p_A = 1 - p_B for all attacks A -> B."""
    def generate_constraints(self):
        for edge in self.af.get_edges():
            var_from = edge.node_from.prob_var
            var_to = edge.node_to.prob_var
            self.constraints.append(var_from == 1 - var_to)


class SemanticsRat(Semantics):
    """Rationality semantics: p_A >= 0.5 implies p_B <= 0.5 for all attacks A -> B."""
    def generate_constraints(self):
        for edge in self.af.get_edges():
            var_from = edge.node_from.prob_var
            var_to = edge.node_to.prob_var
            self.constraints.append(z3.Implies(var_from >= 0.5, var_to <= 0.5, ctx=self.context))


class SemanticsCoh(Semantics):
    """Coherency semantics: p_A <= 1 - p_B for all attacks A -> B."""
    def generate_constraints(self):
        for edge in self.af.get_edges():
            var_from = edge.node_from.prob_var
            var_to = edge.node_to.prob_var
            self.constraints.append(var_from <= 1 - var_to)


class SemanticsOpt(Semantics):
    """Optimism semantics: p_A >= 1 - Sum p_B for all attackers B of A."""
    def generate_constraints(self):
        for node in self.af.get_nodes():
            summation = 0
            for attacker in node.parents:
                summation += attacker.prob_var
            self.constraints.append(node.prob_var >= 1 - summation)


class SemanticsSOpt(Semantics):
    """Semi-optimism semantics: p_A >= 1 - Sum p_B for all attackers B of A iff A has at least one attacker."""
    def generate_constraints(self):
        for node in self.af.get_nodes():
            if node.is_initial():
                continue
            summation = 0
            for attacker in node.parents:
                summation += attacker.prob_var
            self.constraints.append(node.prob_var >= 1 - summation)


class SemanticsJus(Semantics):
    """Justifiability semantics: Coherency and optimism constraints hold."""
    def generate_constraints(self):
        cOpt = SemanticsOpt(self.z3_instance)
        cCoh = SemanticsCoh(self.z3_instance)
        self.constraints.extend(cOpt.get_constraints() + cCoh.get_constraints())


######################
# semantics by Baier
######################

class SemanticsCF(Semantics):
    """Conflict-freeness semantics: p_A_B = 0 for all attacks A -> B."""
    def generate_constraints(self):
        for edge in self.af.get_edges():
            assignment = [(edge.node_from, True), (edge.node_to, True)]
            prob_var = self.z3_instance.get_prob_var(assignment)
            self.constraints.append(prob_var == 0)


# Admissibility

def almost_sure_defense_constraint(z3i: zi.Z3Instance, node: AF.Node):
    """
    Generate a constraint for an argument A to be almost-surely defended.
    This means for all attackers B->A, the probability that B is in turn attacked (and thus A is defended) is 1.
    That is, P(C1 or C2 or ... Cn) = 1 for attackers C1...Cn of B, or, equivalently, P(nC1, nC2, ..., nCn) = 0.
    If an attacker B of A has no attackers, then A cannot be almost-surely defended, i.e. the impossible constraint is
    added.
    """
    constraint = z3.BoolVal(True, ctx=z3i.context)
    for attacker in node.parents:
        if attacker.parents:
            assignment = []
            for defender in attacker.parents:
                assignment.append((defender, False))
            prob_var = z3i.get_prob_var(assignment)
            constraint = z3.And(constraint, prob_var == 0)
        else:
            return z3.BoolVal(False, ctx=z3i.context)
    return constraint


class SemanticsWAdm(Semantics):
    """Weak admissibility semantics: p_A = 1 implies A is almost-surely defended."""
    def generate_constraints(self):
        for node in self.af.get_nodes():
            as_defense_constraint = almost_sure_defense_constraint(self.z3_instance, node)
            self.constraints.append(z3.Implies(node.prob_var == 1, as_defense_constraint, ctx=self.context))


class SemanticsPrAdm(Semantics):
    def generate_constraints(self):
        raise NotImplementedError


class SemanticsMinAdm(Semantics):
    def generate_constraints(self):
        raise NotImplementedError


class SemanticsJntAdm(Semantics):
    def generate_constraints(self):
        raise NotImplementedError


class SemanticsElmAdm(Semantics):
    def generate_constraints(self):
        raise NotImplementedError


# Completeness

class SemanticsWCmp(Semantics):
    """Weak completeness semantics: p_A = 1 if and only if A is almost-surely defended."""
    def generate_constraints(self):
        for node in self.af.get_nodes():
            as_defense_constraint = almost_sure_defense_constraint(self.z3_instance, node)
            self.constraints.append(z3.Implies(node.prob_var == 1, as_defense_constraint, ctx=self.context))
            self.constraints.append(z3.Implies(as_defense_constraint, node.prob_var == 1, ctx=self.context))


class SemanticsPrCmp(Semantics):
    def generate_constraints(self):
        raise NotImplementedError


class SemanticsMinCmp(Semantics):
    def generate_constraints(self):
        raise NotImplementedError


class SemanticsJntCmp(Semantics):
    def generate_constraints(self):
        raise NotImplementedError


class SemanticsElmCmp(Semantics):
    def generate_constraints(self):
        raise NotImplementedError


#####################
# semantics by Käfer
#####################

class SemanticsWNorS(Semantics):
    """
    Weak not-or semantics without CF: Generates constraints p_A <= p_nB1_nB2_..._nBi for every non-initial argument A
    with attackers B1, ..., Bi for some i > 0.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            if node.is_initial():
                continue  # for initial nodes, the constraint would be 'p_A <= 1' which already exists
            assignment = [(attacker, False) for attacker in node.parents]
            attacker_prob_var = self.z3_instance.get_prob_var(assignment)
            self.constraints.append(node.prob_var <= attacker_prob_var)


class SemanticsNorS(Semantics):
    """
    Not-or semantics without CF: Generates constraints p_A = p_nB1_nB2_..._nBi for every non-initial argument A with
    attackers B1, ..., Bi for some i > 0.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            if node.is_initial():
                continue
            assignment = [(attacker, False) for attacker in node.parents]
            attacker_prob_var = self.z3_instance.get_prob_var(assignment)
            self.constraints.append(node.prob_var == attacker_prob_var)


class SemanticsSNorS(Semantics):
    """
    Strong not-or semantics without CF. Generates constraints p_A = 1 for initial arguments and p_A = p_nB1_nB2_..._nBi
    for every non-initial argument A with attackers B1, ..., Bi for some i>0.
    Also the intersection of Nor and Fou.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            if node.is_initial():
                self.constraints.append(node.prob_var == 1)
            else:
                assignment = [(attacker, False) for attacker in node.parents]
                attacker_prob_var = self.z3_instance.get_prob_var(assignment)
                self.constraints.append(node.prob_var == attacker_prob_var)


class SemanticsWNor(Semantics):
    """
    Weak not-or semantics: Generates constraints for conflict-freeness and p_A <= p_nB1_nB2_..._nBi for every
    non-initial argument A with attackers B1, ..., Bi for some i > 0.
    """
    def generate_constraints(self):
        sCF = SemanticsCF(self.z3_instance)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            if node.is_initial():
                continue  # for initial nodes, the constraint would be 'p_A <= 1' which already exists
            assignment = [(attacker, False) for attacker in node.parents]
            attacker_prob_var = self.z3_instance.get_prob_var(assignment)
            self.constraints.append(node.prob_var <= attacker_prob_var)


class SemanticsNor(Semantics):
    """
    Not-or semantics: Generates constraints for conflict-freeness and p_A = p_nB1_nB2_..._nBi for every non-initial
    argument A with attackers B1, ..., Bi for some i > 0.
    """
    def generate_constraints(self):
        sCF = SemanticsCF(self.z3_instance)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            if node.is_initial():
                continue
            assignment = [(attacker, False) for attacker in node.parents]
            attacker_prob_var = self.z3_instance.get_prob_var(assignment)
            self.constraints.append(node.prob_var == attacker_prob_var)


class SemanticsSNor(Semantics):
    """
    Strong not-or semantics: Generates constraints for conflict-freeness, p_A = 1 for initial arguments and
    p_A = p_nB1_nB2_..._nBi for every non-initial argument A with attackers B1, ..., Bi for some i>0.
    Also the intersection of Nor and Fou and CF.
    """
    def generate_constraints(self):
        sCF = SemanticsCF(self.z3_instance)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            if node.is_initial():
                self.constraints.append(node.prob_var == 1)
            else:
                assignment = [(attacker, False) for attacker in node.parents]
                attacker_prob_var = self.z3_instance.get_prob_var(assignment)
                self.constraints.append(node.prob_var == attacker_prob_var)


class SemanticsAF(Semantics):
    """
    Special semantics to add constraints for probability values or intervals specified in the AF input file.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            if node.value is not None:
                assert node.interval is None, "node with both value and interval set"
                self.constraints.append(node.prob_var == node.value)
            elif node.interval is not None:
                low, high = node.interval
                self.constraints.append(node.prob_var >= low)
                self.constraints.append(node.prob_var <= high)


class SemanticsNNor(Semantics):
    """
    Noisy not-or semantics:  TODO
    per default without prior probabilities
    p_A_nB_nC = 1 * p_nB_nC
    p_A_nB_C  = (1 - t) * p_nB_C
    p_A_B_nC  = (1 - s) * p_B_nC
    p_A_B_C   = (1 - s ) * (1 - t) * p_B_C
    """
    def __init__(self, z3i: zi.Z3Instance, use_prior_probs=False):
        self.use_prior_probs = use_prior_probs
        super().__init__(z3i)

    def generate_constraints(self):
        # we use attack strength, so a prob_var should be created for each edge
        self.z3_instance.generate_edge_vars()

        for node in self.af.get_nodes():

            node_prior_prob_var = self.z3_instance.create_real_var("pr_" + node.name)  # e.g. pr_A
            if self.use_prior_probs:
                if node.value is not None:
                    self.constraints.append(node_prior_prob_var == node.value)
                elif node.interval is not None:
                    i_min, i_max = node.interval
                    self.constraints.append(z3.And(node_prior_prob_var >= i_min, node_prior_prob_var <= i_max))
                else:
                    self.constraints.append(z3.And(node_prior_prob_var >= 0, node_prior_prob_var <= 1))
            else:
                self.constraints.append(node_prior_prob_var == 1)

            if node.is_initial():
                self.constraints.append(node.prob_var == node_prior_prob_var)
                continue

            for parent_assignment in zi.assignments(node.parents):
                parent_prob_var = self.z3_instance.get_prob_var(parent_assignment)  # e.g. p_nB_C
                prob_var = self.z3_instance.get_prob_var(parent_assignment + [(node, True)])  # e.g. p_A_nB_C

                product = node_prior_prob_var * parent_prob_var
                for (parent, status) in parent_assignment:
                    if status:
                        edge: AF.Edge = node.get_parent_edge(parent)
                        product *= (1 - edge.prob_var)
                self.constraints.append(prob_var == product)


class SemanticsNNorAF(SemanticsNNor):
    """
    Noisy not-or semantics:  TODO
    with prior probabilities
    p_A_nB_nC = prA * p_nB_nC
    p_A_nB_C  = prA * (1 - t) * p_nB_C
    p_A_B_nC  = prA * (1 - s) * p_B_nC
    p_A_B_C   = prA * (1 - s ) * (1 - t) * p_B_C
    """
    def __init__(self, z3i: zi.Z3Instance):
        super().__init__(z3i, use_prior_probs=True)


class SemanticsCFs(Semantics):
    """Conflict-freeness semantics with attack strengths: p_A_B <= 1-s for all attacks A -s-> B."""
    def generate_constraints(self):
        # we use attack strength, so a prob_var should be created for each edge
        self.z3_instance.generate_edge_vars()

        for edge in self.af.get_edges():
            assignment = [(edge.node_from, True), (edge.node_to, True)]
            prob_var = self.z3_instance.get_prob_var(assignment)
            self.constraints.append(prob_var <= 1 - edge.prob_var)


###################
# Helper functions
###################

def get_semantics_class_by_name(name: str):
    """
    By dark magic, get the semantics class given its name as string, e.g. "wNor" to yield the class 'SemanticsWNor'.

    :param name: the short name of the semantics (case insensitive)
    :return: The semantics class corresponding to the name
    """
    name = "semantics" + name.lower()
    queue = Semantics.__subclasses__().copy()
    while queue:
        semantics_class = queue.pop()
        if name == semantics_class.__name__.lower():
            return semantics_class
        queue.extend(semantics_class.__subclasses__())
    print("Error: No semantics called 'name' was found.")
    return None


def all_semantics_short_names():
    """
    :return: A list of the short names of all semantics declared in this file
    """
    names = []
    queue = Semantics.__subclasses__().copy()
    while queue:
        semantics_class = queue.pop()
        queue.extend(semantics_class.__subclasses__())
        name = semantics_class.__name__
        if name[0:9] == "Semantics":
            name = name[9:]  # strip away leading 'Semantics'
        names.append(name)
    return names



from typing import Set, Tuple, Dict


class Node:
    """A node in an abstract argumentation framework, usually a single argument."""
    __default_node_prefix = "A"

    def __init__(self, node_id, name: str = ""):
        self.id = node_id
        self.parents: Set[Node] = set()
        self.parents_edges = dict()  # Dict[Node.id, Edge]
        self.children: Set[Node] = set()
        self.children_edges = dict()  # Dict[Node.id, Edge]
        self.value = None  # float
        self.interval = None  # Tuple[float, float]
        self.prob_var = None  # z3.ArithRef
        if not name:
            self.name = self.__default_node_prefix + str(self.id)
        else:
            self.name = name

    def add_value(self, value: float):
        self.value = value

    def add_interval(self, interval: Tuple[float, float]):
        self.interval = interval

    def is_initial(self):
        return len(self.parents) == 0

    def get_parent_edge(self, node_from):
        if node_from.id in self.parents_edges:
            return self.parents_edges[node_from.id]
        else:
            raise ValueError("Node '" + self.name + "' has no parent edge from node '" + node_from.name + "'.")

    def get_child_edge(self, node_to):
        if node_to.id in self.children_edges:
            return self.children_edges[node_to.id]
        else:
            raise ValueError("Node '" + self.name + "' has no child edge to node '" + node_to.name + "'.")

    def __eq__(self, other):
        return self.id == other.id

    def __lt__(self, other):
        return self.id < other.id

    def __gt__(self, other):
        return self.id > other.id

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return self.name
        # return "Node(" + self.id + ": " + self.name + ")"


class Edge:
    """A directed edge in an abstract argumentation framework, usually an attack."""
    __default_label_prefix = "s"  # default label "s_A_B" for an attack from node A to node B

    def __init__(self, node_from, node_to, label=""):
        self.node_from: Node = node_from
        self.node_to: Node = node_to
        self.label: str = label
        self.value = None
        self.interval = None
        self.prob_var = None  # z3.ArithRef
        if self.label == "":
            self.label = self.__default_label_prefix + '_' + self.node_from.name + '_' + self.node_to.name
        self.node_from.children.add(self.node_to)
        self.node_from.children_edges[node_to.id] = self
        self.node_to.parents.add(self.node_from)
        self.node_to.parents_edges[node_from.id] = self

    def add_value(self, value: float):
        self.value = value

    def add_interval(self, interval: Tuple[float, float]):
        self.interval = interval


class AF:
    """A class representing an abstract argumentation framework."""
    def __init__(self, name: str = ""):
        self.name = name
        self.nodes: Dict[str, Node] = dict()
        self.edges: Set[Edge] = set()

    def add_node(self, node: Node):
        self.nodes[node.id] = node

    def add_edge(self, edge: Edge):
        self.edges.add(edge)

    def get_node(self, node_id: str):
        return self.nodes[node_id]

    def get_nodes(self):
        return self.nodes.values()

    def get_node_by_name(self, name: str):
        for node in self.get_nodes():
            if node.name == name:
                return node
        raise ValueError("No node with name '" + name + "' found.")

    def get_initial_nodes(self):
        return set([node for node in self.nodes.values() if node.is_initial()])

    def get_edges(self):
        return self.edges

    def print(self):
        print("AF " + self.name)
        print(str(len(self.nodes)) + " nodes: " + ', '.join([node.name for node in self.nodes.values()]))
        print(str(len(self.edges)) + " edges: "
              + ', '.join([edge.node_from.name + '->' + edge.node_to.name for edge in self.edges]))



import AF


def parse_interval(string, obj):
    """Parse a value or an interval and add it to either the node or edge passed as obj.
    '0.2:0.8' -> add interval (0.2, 0.8)
    '0.5' -> add value 0.5
    'ABC' -> print warning, add nothing"""

    split = string.split(':')

    p1 = float(split[0])
    if len(split) == 1:
        obj.add_value(p1)
    elif len(split) == 2:
        p2 = float(split[1])
        obj.add_interval((p1, p2))
    else:
        print("Warning: Ill-formed probability '" + string + "'")


def parse_tgf(path):
    """Parse a file in trivial graph format. Return an AF object."""
    with open(path, "r") as file:
        tgf_file = file.read()

    af = AF.AF(path)

    # remove blank lines and line comments (starting with ';')
    lines = tgf_file.split('\n')
    node_declarations = []
    edge_declarations = []
    start_edges = False
    for line in lines:

        split = line.split(";")
        if split[0].strip() == "#":
            start_edges = True
            continue
        if split[0].strip() != "":
            if start_edges:
                edge_declarations.append(split[0])
            else:
                node_declarations.append(split[0])

    # parse nodes and edges
    for d in node_declarations:
        # format: <node_id> [<node_name>] [<node_prob>|<node_prob_min>:<node_prob_max>]
        # if d.strip() == "":
        #     continue
        tokens = [token for token in d.split(' ') if token.strip()]
        node_id = tokens[0]
        node = AF.Node(node_id)

        if len(tokens) > 1:
            next_token = tokens[1]
            if next_token[0].isalpha():  # <node_name> is given
                node.name = next_token
                if len(tokens) > 2:
                    next_token = tokens[2]
                else:
                    next_token = None
            if next_token:
                parse_interval(next_token, node)
        af.add_node(node)

    for d in edge_declarations:
        # if d.strip() == "":
        #     continue

        tokens = [token for token in d.split(' ') if token.strip()]
        node_from_id = tokens[0]
        node_from = af.get_node(node_from_id)
        node_to_id = tokens[1]
        node_to = af.get_node(node_to_id)
        edge = AF.Edge(node_from, node_to)
        if len(tokens) > 2:
            next_token = tokens[2]
            if next_token[0].isalpha():
                edge.label = next_token
                if len(tokens) > 3:
                    next_token = tokens[3]
                else:
                    next_token = None
            if next_token:
                parse_interval(next_token, edge)
        af.add_edge(edge)

    af.print()
    return af



from typing import Dict, List, Tuple, Any

import AF
import z3


class Z3Instance:
    """Class to bundle the declared probabilistic variables, the underlying full joint distribution."""

    def __init__(self, af: AF.AF):
        self.af = af
        self.context = z3.Context()
        self.solver = z3.Solver(ctx=self.context)
        self.full_joint_prob_vars: Dict[str, z3.ArithRef] = dict()
        self.other_prob_vars: Dict[str, z3.ArithRef] = dict()
        self.real_vars: Dict[str, z3.ArithRef] = dict()
        self.constraints = set()
        self.af_constraints = set()

        self._generate_full_prob_vars()

        # Add a prob_var for each node in the AF
        for node in self.af.get_nodes():
            node.prob_var = self.get_prob_var([(node, True)])
            # if node.value is not None:
            #     self.af_constraints.add(node.prob_var == node.value)
            #     assert node.interval is None, "both node.value and node.interval where given"
            # elif node.interval is not None:
            #     (val_min, val_max) = node.interval
            #     self.af_constraints.add(z3.And(node.prob_var >= val_min, node.prob_var <= val_max))

    def generate_edge_vars(self):
        """
        Generates a probabilistic variable for each edge. If an edge has a value or interval given, those are added as
        constraints to af_constraints.
        """
        for edge in self.af.get_edges():
            if edge.prob_var is not None:
                # There might be several calls to generate_edge_vars, but we want to add the constraints only once.
                continue
            edge.prob_var = z3.Real(edge.label, self.context)

            if edge.value is not None:
                self.af_constraints.add(edge.prob_var == edge.value)
                assert edge.interval is None, "both edge.value and edge.interval where given"
            elif edge.interval is not None:
                (val_min, val_max) = edge.interval
                self.af_constraints.add(z3.And(edge.prob_var >= val_min, edge.prob_var <= val_max))
            else:
                self.af_constraints.add(edge.prob_var >= 0)
                self.af_constraints.add(edge.prob_var <= 1)

    def _generate_full_prob_vars(self):
        """Adds variables for the full joint distribution to the context. Adds the constraint that they sum to one."""
        constraint = 0
        for assignment in assignments(self.af.get_nodes()):
            assignment_prob_name = prob_name(assignment)

            prob_var = z3.Real(assignment_prob_name, self.context)
            self.full_joint_prob_vars[assignment_prob_name] = prob_var
            self.constraints.add(prob_var >= 0)
            self.constraints.add(prob_var <= 1)
            constraint += prob_var

        constraint = 1 == constraint
        self.constraints.add(constraint)

    def get_prob_var(self, assignment: List[AF.Tuple[AF.Node, bool]]):
        """
        Get the probabilistic variable corresponding to a full or partial assignment.
        If it does not exist yet, it is created along with the constraints 'new_var >= 0', 'new_var <= 1' and the
        representation of new_var as sum of the full joint variables.
        """
        assert assignment, "empty assignment given"
        prob_var_name = prob_name(assignment)
        if prob_var_name in self.other_prob_vars:
            prob_var = self.other_prob_vars[prob_var_name]
        else:
            prob_var = z3.Real(prob_var_name, self.context)
            self.other_prob_vars[prob_var_name] = prob_var
            self.constraints.add(prob_var >= 0)
            self.constraints.add(prob_var <= 1)
            self.constraints.add(self.sum_out(assignment, prob_var))
        return prob_var

    def create_real_var(self, name: str):
        """
        Creates and returns an auxiliary z3 real variable.
        """
        if name in self.real_vars:
            raise ValueError("A real var with name '" + name + "' already exists.")
        real_var = z3.Real(name, self.context)
        self.real_vars[name] = real_var
        return real_var

    def sum_out(self, part_assignment: List[AF.Tuple[AF.Node, bool]], part_assignment_prob_var):
        """[(A,True),(B,False)], [A,B,C] -> p_A_nB = p_A_nB_C + p_A_nB_nC"""
        part_assignment_args = [arg for (arg, _) in part_assignment]
        remaining_args = [arg for arg in self.af.get_nodes() if arg not in part_assignment_args]
        summation = 0
        for assignment in assignments(remaining_args):

            full_assignment = part_assignment + assignment
            full_assignment_name = prob_name(full_assignment)
            prob_var = self.full_joint_prob_vars[full_assignment_name]
            summation += prob_var
        constraint = part_assignment_prob_var == summation
        return constraint
    
    def add_constraints(self, constraints):
        for cons in constraints:
            self.constraints.add(cons)

    def run(self, prob_constraints):
        """
        TODO

        :param prob_constraints: a list of constraints
        :return: A z3 CheckSatResult, i.e. sat, unsat or unknown
        """
        self.solver.reset()
        self.solver.add(self.constraints)
        # print("self.constraints:", self.constraints)
        self.solver.add(self.af_constraints)
        # print("self.af_constraints", self.af_constraints)
        self.solver.add(prob_constraints)
        # print("prob_constraints", prob_constraints)

        result = self.solver.check()
        # print(result)
        # if result == z3.sat:
        #     model = self.solver.model()
        #     print(model)
        #     for node in self.af.get_nodes():
        #         prob = model.eval(node.prob_var)
        #         print(node.name, prob)
        return result

    def print_distribution(self, model: z3.ModelRef, nodes=None, only_positive=False):
        """
        Print distribution given by the model over the assignments of the given nodes.
        If no list of nodes is given, the distributions over all nodes in the AF is given.
        """
        if not nodes:
            nodes = self.af.get_nodes()
        node_assignments = assignments(nodes)
        for assignment in node_assignments:
            prob_var = self.get_prob_var(assignment)
            value = get_prob_var_value(prob_var, model)
            if not only_positive or value > 0:
                print(prob_var, "=", value)


def prob_name(argument_combo: List[AF.Tuple[AF.Node, bool]]) -> str:
    """
    Expects a list of tuples of nodes matched to True or False, i.e. [(node_B, False), (node_C, True)].
    Generates a name for the probability of the conjunction of the nodes, i.e. 'p_nB_C'.
    The order is determined by node IDs.
    """
    name = "p"
    argument_combo.sort()
    for (arg, b) in argument_combo:
        name += "_"
        if not b:
            name += "n"
        name += arg.name
    return name


def assignments(ar: list, br: list = None) -> List[List[Tuple[Any, Any]]]:
    """Generates all possible assignments of the elements in list 'ar' to elements of list 'br'.
     E.g. ar = [x,y], br = [0,1] -> [[(x,0),(y,0)],[(x,0),(y,1)],[(x,1),(y,0)],[(x,1),(y,1)]]"""
    if br is None:
        br = [False, True]
    combinations = [[]]

    for a in ar:
        next_combinations = []
        for c in combinations:
            for b in br:
                new_c = c.copy()
                new_c.append((a, b))
                next_combinations.append(new_c)
        combinations = next_combinations
    return combinations


def get_prob_var_value(prob_var, model: z3.ModelRef):
    """
    Get the value assigned to a probabilistic variable by the given model.
    :return: The value as float
    """
    # Based on https://stackoverflow.com/a/12600208/6620204
    value = model.eval(prob_var)
    if z3.is_int_value(value):
        return value.as_long()
    elif z3.is_rational_value(value):
        return float(value.numerator_as_long())/float(value.denominator_as_long())
    elif z3.is_algebraic_value(value):
        approx_value = value.approx(20)
        return float(approx_value.numerator_as_long())/float(approx_value.denominator_as_long())
    else:
        raise ValueError("Unable to convert to float: " + str(value))


def print_model(af: AF.AF, model: z3.ModelRef):
    for node in af.get_nodes():
        print(node.name, "=", get_node_value(node, model))


def get_node_value(node: AF.Node, model: z3.ModelRef):
    """
    Get the value assigned to a node's probabilistic variable by the given model.
    :return: The value as float
    """
    return get_prob_var_value(node.prob_var, model)
